@@ -5,4 +5,7 @@ from hcai_datasets import hcai_faces
 '''Audio Datasets'''
 from hcai_datasets import hcai_audioset
 from hcai_datasets import hcai_librispeech
-from hcai_datasets import hcai_is2021_ess
\ No newline at end of file
+from hcai_datasets import hcai_is2021_ess
+
+'''Nova Datasets'''
+from hcai_datasets import hcai_nova_dynamic
\ No newline at end of file
